@@ -1,7 +1,9 @@
-
 /*
  * GNU Typist  - interactive typing tutor program for UNIX systems
- * Copyright (C) 1998, 1999, 2000, 2001, 2002  Simon Baldwin (simonb@sco.com)
+ * 
+ * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003  
+ * 					Simon Baldwin (simonb@sco.com)
+ * Copyright (C) 2003  GNU Typist Development Team <bug-gtypist@gnu.org>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -16,52 +18,44 @@
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
  */
 
-
+#include "config.h"
 #include <stdlib.h>
 #include <unistd.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <signal.h>
-/* necessary on SuSE there are no ncurses* -> curses* links */
+
 #ifdef HAVE_LIBCURSES
 #include <curses.h>
 #else
-#include <ncurses.h>
+#include <ncurses/curses.h>
 #endif
+
 #include <time.h>
 #include <errno.h>
 #include <string.h>
 #include <getopt.h>
 #include <assert.h>
-#include "config.h"
+#include <locale.h>
 
-/* Internationalization */
+#include "cursmenu.h"
+#include "script.h"
+#include "error.h"
+#include "gtypist.h"
 
-#ifndef ENABLE_NLS
-
-#define _(String) (String)
-
-#else /* not ENABLE_NLS */
-
-#include <libintl.h>
+#include "gettext.h"
 #define _(String) gettext (String)
 
-#endif /* not ENABLE_NLS */
-
 /* VERSION and PACKAGE defined in config.h */
 
 char *COPYRIGHT;
 
 /* a definition of a boolean type */
+#ifndef bool
 #define bool			int
-
-/* limits on line lengths from the script file and screen */
-#define	MAX_SCR_LINE		1024
-#define	MIN_SCR_LINE		2
-#define	MAX_WIN_LINE		128
+#endif
 
 /* some screen postions */
 #define	MESSAGE_LINE		(LINES - 1)
@@ -71,30 +65,6 @@ char *COPYRIGHT;
 #define DP_TOP_LINE		(I_TOP_LINE + 2)
 #define	SPEED_LINE		(LINES - 5)
 
-/* things to help parse the input file */
-#define	SCR_COMMAND(X)		(X[0])
-#define	SCR_SEP(X)		(X[1])
-#define	SCR_DATA(X)		(&X[2])
-#define	C_COMMENT		'#'
-#define	C_ALT_COMMENT		'!'
-#define	C_SEP			':'
-#define	C_CONT			' '
-#define	C_LABEL			'*'
-#define	C_TUTORIAL		'T'
-#define	C_INSTRUCTION		'I'
-#define	C_CLEAR			'B'
-#define	C_GOTO			'G'
-#define	C_EXIT			'X'
-#define	C_QUERY			'Q'
-#define	C_YGOTO			'Y'
-#define	C_NGOTO			'N'
-#define	C_DRILL			'D'
-#define C_DRILL_PRACTICE_ONLY   'd'
-#define	C_SPEEDTEST		'S'
-#define	C_SPEEDTEST_PRACTICE_ONLY 's'
-#define	C_KEYBIND		'K'
-#define C_ERROR_MAX_SET         'E'
-#define C_ON_FAILURE_SET        'F'
 
 /* mode indicator strings */
 char *MODE_TUTORIAL;
@@ -110,21 +80,6 @@ char *MODE_SPEEDTEST;
 /* this is needed in wait_user
    DJGPP defines '\n' as 0x0A, but pdcurses 2.4 returns 0x0D (because
    raw() is called) */
-#ifndef DJGPP
-#define ASCII_ENTER             '\n'
-#else
-#define ASCII_ENTER             0x0D
-#endif
-#define	ASCII_NL		'\n'
-#define	ASCII_NULL		'\0'
-#define	ASCII_ESC		27
-#define	ASCII_BELL		7
-#define	ASCII_SPACE		' '
-#define	ASCII_DASH		'-'
-#define	ASCII_TAB		'\t'
-#define	ASCII_BS		8
-#define	ASCII_DEL		127
-#define	STRING_NL		"\n"
 char *WAIT_MESSAGE;
 char *ERROR_TOO_HIGH_MSG;
 char *SKIPBACK_VIA_F_MSG;
@@ -135,8 +90,8 @@ char *NO_SKIP_MSG;
 char *SPEED_RAW;
 char *SPEED_ADJ;
 char *SPEED_PCT_ERROR;
-char *YN;
-char *RNE;
+unsigned char *YN;
+unsigned char *RNE;
 
 #ifndef DATADIR
 #define DATADIR "."
@@ -150,7 +105,6 @@ static short	colour_array[] = {
   COLOR_BLACK, COLOR_RED, COLOR_GREEN, COLOR_YELLOW,
   COLOR_BLUE, COLOR_MAGENTA, COLOR_CYAN, COLOR_WHITE };
 #define	NUM_COLOURS		(sizeof( colour_array ) / sizeof( short ))
-#define	C_NORMAL		1		/* normal colour pair */
 
 /* shortcuts for reverse/normal mode strings */
 #define	ADDSTR_REV(X)		do { attron( A_REVERSE ); addstr( X ); \
@@ -171,15 +125,19 @@ static int	cl_curs_flash = 10;		/* cursor flash period */
 static bool	cl_silent = FALSE;		/* no beep on errors */
 static char	*cl_start_label = NULL;		/* initial lesson start point*/
 static bool	cl_colour = FALSE;		/* set if -c given */
-static int	cl_fgcolour = 0;		/* fg colour */
+static int	cl_fgcolour = 7;		/* fg colour */
 static int	cl_bgcolour = 0;		/* bg colour */
+static int	cl_banner_bg_colour = 0;	// since we display them in
+static int	cl_banner_fg_colour = 6;	// inverse video, fg is bg
+static int	cl_prog_name_colour = 5;	// and vice versa.
+static int 	cl_prog_version_colour = 1;
+static int	cl_menu_title_colour = 7;
 static bool	cl_wp_emu = FALSE;		/* do wp-like stuff */
 static bool     cl_no_skip = FALSE;             /* forbid the user to */
+static bool	cl_rev_video_errors = FALSE;    /* reverse video for errors */
 
 /* a few global variables */
-static char	*argv0 = NULL;
 static bool	global_resp_flag = TRUE;
-static int	global_line_counter = 0;
 static char	global_prior_command = C_CONT;
 
 static float    global_error_max = -1.0;
@@ -188,20 +146,9 @@ static bool     global_error_max_persistent = FALSE;
 static struct label_entry *global_on_failure_label = NULL;
 static bool     global_on_failure_label_persistent = FALSE;
 
-/* a global area for label indexing - singly linked lists, hashed */
-#define	NLHASH			32		/* num hash lists */
-struct label_entry {
-  char		*label;			/* label string */
-  long		offset;			/* offset into file */
-  int		line_count;		/* line number in script */
-  struct label_entry	*next;		/* pointer to next element */
-};
-static struct label_entry	*global_label_list[NLHASH];
-/* list head */
-
 /* a global area for associating function keys with labels */
 #define NFKEYS			12		/* num of function keys */
-static	char	*fkey_bindings[ NFKEYS ] =
+static char	*fkey_bindings[ NFKEYS ] =
   { NULL, NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL, NULL };
 /* table of pseudo-function keys, to allow ^Q to double as Fkey1, etc */
@@ -215,25 +162,21 @@ static	char	pfkeys[ NFKEYS ] =
 #define MAX(A,B) ((A)<(B)?(B):(A))
 
 
+static bool user_is_always_sure = FALSE;
+
 /* prototypes */
 
 static int getch_fl( char cursor_char );
-static void fatal_error( char *message, char *line );
-static void get_script_line( FILE *script, char *line );
-static int hash_label( char *label );
 static void index_labels( FILE *script );
-static void seek_label( FILE *script, char *label, char *ref_line );
-static void wait_user( char *message, char *mode );
+static void wait_user (FILE *script, char *message, char *mode );
 static void display_speed( int total_chars, long elapsed_time, int errcount );
 static void do_keybind( FILE *script, char *line );
 static void do_tutorial( FILE *script, char *line );
 static void do_instruction( FILE *script, char *line );
-static char *buffer_command( FILE *script, char *line );
 static void do_drill( FILE *script, char *line );
 static void do_speedtest( FILE *script, char *line );
 static void do_clear( FILE *script, char *line );
 static void do_goto( FILE *script, char *line, bool flag );
-static void do_exit( FILE *script );
 static char do_query_repeat( FILE *script, bool allow_next );
 static bool do_query_simple( char *text );
 static bool do_query( FILE *script, char *line );
@@ -248,6 +191,82 @@ static void print_help();
 static void parse_cmdline( int argc, char **argv );
 static void catcher( int signal );
 
+// Display the top banner with the given text
+void banner (const char *text)
+{
+   int colnum, brand_length, brand_position, text_length, text_position;
+   	
+   // Get rid of spaces at the edges of the text
+   while (isspace (*text))
+      text ++;
+   text_length = strlen (text);
+   if (text_length > 0)
+      while (isspace (text [text_length - 1]))
+      {
+         text_length --;
+	 if (! text_length)
+	    break;
+      }
+
+   brand_length = strlen (PACKAGE) + strlen (VERSION) + 3,
+   brand_position = COLS - brand_length,
+   text_position = ((COLS - brand_length) > text_length) ?
+		(COLS - brand_length - text_length) / 2 : 0;
+	
+// TODO:  much of redundant output here...
+
+   move (B_TOP_LINE , 0);
+   attron (COLOR_PAIR (C_BANNER));
+   for (colnum = 0; colnum < COLS; colnum++)
+      ADDCH_REV (ASCII_SPACE);
+   
+   move (B_TOP_LINE, text_position);
+   {
+      int i;
+      for (i = 0; i < text_length; i ++)
+         ADDCH_REV (text [i]);
+   }
+
+   move (B_TOP_LINE, brand_position);
+   attron (COLOR_PAIR (C_PROG_NAME));
+   ADDCH_REV (' ');
+   ADDSTR_REV (PACKAGE);
+   ADDCH_REV (' ');
+   attron (COLOR_PAIR (C_PROG_VERSION));
+   ADDSTR_REV (VERSION);
+   refresh ();
+   attron (COLOR_PAIR (C_NORMAL));
+}
+
+void bind_F12 (const char *label)
+{
+  if (!label)
+     return;
+  
+  if (fkey_bindings [11])
+     free (fkey_bindings [11]);
+  fkey_bindings [11] = strdup (label);
+  if (! fkey_bindings [11])
+  {
+       perror ("strdup");
+       fatal_error (_("internal error in strdup"), label);
+  }
+}
+
+// Return to the last F12-binded location
+static inline void escape (FILE *script)
+{
+   if (fkey_bindings [11])
+   {
+      if (*(fkey_bindings [11]))
+         seek_label (script, fkey_bindings [11], NULL); 
+      else
+	 do_exit (script);
+   }
+   else
+      do_exit (script);
+}
+
 /*
   getch() that does a flashing cursor; some xterms seem to be
   unwilling to do A_BLINK, however, the program needs some
@@ -317,85 +336,6 @@ getch_fl( char cursor_char )
   return ( return_char );
 }
 
-
-/*
-  handle fatal errors (pretty much any error) by dropping out
-  of curses etc, and printing a complaint
-  message that is already translated to the appropriate language 
-*/
-static void 
-fatal_error( char *message, char *line ) {
-
-  /* stop curses */
-  if ( cl_colour && has_colors() )
-    wbkgdset( stdscr, 0 );
-  clear(); refresh(); endwin();
-  
-  /* print out the error message and stop */
-  fprintf( stderr, "%s: %s %d: %s", argv0, _("line"), global_line_counter,
-	   message );
-  if ( line != NULL )
-    fprintf( stderr, ":\n%s\n", line );
-  else
-    fprintf( stderr, "\n" );
-  exit( 1 );
-}
-
-
-/*
-  get the next non-comment, non-blank line from the script file
-  and check its basic format
-*/
-static void 
-get_script_line( FILE *script, char *line ) {
-
-  /* get lines until not empty/comment, or eof found */
-  fgets(line, MAX_SCR_LINE, script);
-  global_line_counter++;
-  while ( ! feof( script ) && 
-	  ( strcmp( line, "\n" ) == 0 ||
-	    SCR_COMMAND( line ) == C_COMMENT ||
-	    SCR_COMMAND( line ) == C_ALT_COMMENT )) 
-    {
-      fgets(line, MAX_SCR_LINE, script);
-      global_line_counter++;
-    }
-
-  /* if a line was read then check it */
-  if ( ! feof( script )) 
-    {
-      if ( line[strlen( line ) - 1] == ASCII_NL )
-	line[strlen( line ) - 1] = ASCII_NULL;
-      if ( strlen( line ) < MIN_SCR_LINE )
-	fatal_error( _("data shortage"), line );
-      if ( SCR_SEP( line ) != C_SEP )
-	fatal_error( _("missing ':'"), line );
-      if ( SCR_COMMAND( line ) != C_LABEL 
-	   && SCR_COMMAND( line ) != C_GOTO 
-	   && SCR_COMMAND( line ) != C_YGOTO
-	   && SCR_COMMAND( line ) != C_NGOTO
-	   && strlen( SCR_DATA( line )) > COLS )
-	fatal_error( _("line too long for screen"), line );
-    }
-}
-
-
-/*
-  label hashing function - returns an index for the lists
-*/
-static int 
-hash_label( char *label ) {
-  char	*p;				/* pointer through string */
-  int	csum = 0;			/* sum of characters */
-  
-  /* hash by summing the characters and taking modulo of the
-     number of hash lists defined */
-  for ( p = label; *p != ASCII_NULL; p++ )
-    csum += *p;
-  return ( csum % NLHASH );
-}
-
-
 /*
   go through the file and index all the labels we can find
 */
@@ -479,49 +419,10 @@ static void index_labels( FILE *script ) {
     }
 }
 
-
-/*
-  search out a label from the file, and set the file pointer to
-  that location
-*/
-static void 
-seek_label( FILE *script, char *label, char *ref_line ) {
-  struct label_entry	*check_label;	/* pointer through list */
-  int	hash;				/* hash index */
-  char	err[MAX_SCR_LINE];		/* error message string */
-  
-  /* find the right hash list for the label */
-  hash = hash_label( label );
-  
-  /* search the linked list for the label */
-  for ( check_label = global_label_list[ hash ]; check_label != NULL;
-	check_label = check_label->next ) 
-    {
-      
-      /* see if this is our label */
-      if ( strcmp( check_label->label, label ) == 0 )
-	break;
-    }
-
-  /* see if the label was not found in the file */
-  if ( check_label == NULL ) 
-    {
-      sprintf( err, _("label '%s' not found"), label );
-      fatal_error( err, ref_line );
-    }
-
-  /* move to the label position in the file */
-  if ( fseek( script, check_label->offset, SEEK_SET ) == -1 )
-    fatal_error( _("internal error: fseek"), ref_line );
-  global_line_counter = check_label->line_count;
-}
-
-
 /*
   wait for a nod from the user before continuing
 */
-static void
-wait_user( char *message, char *mode )
+static void wait_user (FILE *script, char *message, char *mode)
 {
   int	resp;			/* response character */
 
@@ -533,8 +434,13 @@ wait_user( char *message, char *mode )
   ADDSTR_REV( message );
 
   do {
-    resp = getch_fl( ASCII_NULL );
-  } while (resp != ASCII_ENTER);
+    resp = getch_fl (ASCII_NULL);
+    if (resp == ASCII_ESC)
+    {
+       escape (script);
+       break;
+    }
+  } while (resp != ASCII_ENTER && resp != ASCII_SPACE);
 
   /* clear the message line */
   move( MESSAGE_LINE, 0 ); clrtoeol();
@@ -642,7 +548,7 @@ do_tutorial( FILE *script, char *line ) {
   /* wait for a return, unless the next command is a query,
      when we can skip it to save the user keystrokes */
   if ( SCR_COMMAND( line ) != C_QUERY )
-    wait_user( WAIT_MESSAGE, MODE_TUTORIAL );
+    wait_user (script, WAIT_MESSAGE, MODE_TUTORIAL);
   global_prior_command = C_TUTORIAL;
 }
 
@@ -672,46 +578,6 @@ do_instruction( FILE *script, char *line ) {
   global_prior_command = C_INSTRUCTION;
 }
 
-
-/*
-  buffer up the complete data from a command; used for D and P
-*/
-static char *buffer_command( FILE *script, char *line ) {
-  int	total_chars = 0;		/* character counter */
-  char	*data = NULL;			/* data string */
-
-  /* get the complete exercise into a single string */
-  do 
-    {
-      
-      /* allocate space for the extra data */
-      if ( total_chars == 0 )
-	data = (char*)malloc( strlen(SCR_DATA( line )) +
-			      strlen(STRING_NL) + 1 );
-      else
-	data = (char*)realloc( data, strlen( data ) +
-			       strlen(SCR_DATA( line )) +
-			       strlen(STRING_NL) + 1 );
-      if ( data == NULL )
-	fatal_error( _("internal error: malloc"), line );
-      
-      /* store the data in the allocated area */
-      if ( total_chars == 0 )
-	strcpy( data, "" );
-      strcat( data, SCR_DATA( line ) );
-      strcat( data, STRING_NL );
-      total_chars = strlen( data );
-      
-      /* and get the next script line */
-      get_script_line( script, line );
-    } 
-  while ( SCR_COMMAND( line ) == C_CONT && ! feof( script ));
-
-  /* return our (malloced) data */
-  return( data );
-}
-
-
 /*
   execute a typing drill
 */
@@ -729,6 +595,7 @@ do_drill( FILE *script, char *line ) {
   char	drill_type;		 /* note of the drill type */
   int	chars_typed;		 /* count of chars typed */
   bool  seek_done = FALSE;       /* was there a seek_label before exit? */
+  int	error_sync;		 /* error resync state */
 
   /* note the drill type to see if we need to make the user repeat */
   drill_type = SCR_COMMAND( line );
@@ -775,15 +642,19 @@ do_drill( FILE *script, char *line ) {
       move( linenum, 0 );
       for ( p = data; *p == ASCII_SPACE && *p != ASCII_NULL; p++ )
 	ADDCH( *p );
-      for ( chars_typed = 0, errors = 0; *p != ASCII_NULL; p++ ) 
+
+      for ( chars_typed = 0, errors = 0, error_sync = 0;
+	    *p != ASCII_NULL; p++ ) 
 	{
 	  rc = getch_fl( *p == ASCII_TAB ?
 			 ASCII_TAB : ASCII_SPACE );
 	  c = (char)rc;
 	  /* this is necessary for DOS: when using raw(), pdcurses 2.4's
 	     getch() returns 0x0D on DOS/Windows  */
+#ifdef DJGPP
 	  if ( c == 0x0D )
 	    rc = c = 0x0A;
+#endif
 
 	  while ( rc == KEY_BACKSPACE
 		  || c == ASCII_BS || c == ASCII_DEL ) 
@@ -797,6 +668,7 @@ do_drill( FILE *script, char *line ) {
 	  if ( chars_typed == 0 )
 	    start_time = (long)time( NULL );
 	  chars_typed++;
+	  error_sync--;
 	
 	  /* ignore delete or backspace in drills */
 	  if ( rc == KEY_BACKSPACE ||
@@ -818,14 +690,30 @@ do_drill( FILE *script, char *line ) {
 	    ADDCH( c );
 	  else 
 	    {
+	      /* try to sync with typist behind */
+	      if ( error_sync >= 0 && p > data && c == *(p-1) )
+		{
+		  p--;
+		  continue;
+		}
+
 	      ADDCH_REV( *p == ASCII_NL ? DRILL_NL_ERR :
 			 (*p == ASCII_TAB ?
-			  ASCII_TAB : DRILL_CH_ERR ));
+			  ASCII_TAB : (cl_rev_video_errors ?
+				       c : DRILL_CH_ERR)));
 	      if ( ! cl_silent ) 
 		{
 		  putchar( ASCII_BELL ); fflush( stdout );
 		}
 	      errors++;
+	      error_sync = 1;
+
+	      /* try to sync with typist ahead */
+	      if ( c == *(p+1) )
+		{
+		  ungetch( c );
+		  error_sync++;
+		}
 	    }
 	
 	  /* move screen location if newline */
@@ -890,7 +778,7 @@ do_drill( FILE *script, char *line ) {
 	      (float)errors/(float)chars_typed > global_error_max/100.0) 
 	    {
 	      sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
-	      wait_user( message, MODE_DRILL );
+	      wait_user (script, message, MODE_DRILL);
 
 	      /* check for F-command */
 	      if (global_on_failure_label != NULL)
@@ -905,7 +793,7 @@ do_drill( FILE *script, char *line ) {
 		  /* reset value unless persistent */
 		  if (!global_on_failure_label_persistent)
 		    global_on_failure_label = NULL;
-		  wait_user( message, MODE_DRILL );
+		  wait_user (script, message, MODE_DRILL);
 		  seek_done = TRUE;
 		  break;
 		}
@@ -960,6 +848,7 @@ do_speedtest( FILE *script, char *line ) {
   char	drill_type;		 /* note of the drill type */
   int	chars_typed;		 /* count of chars typed */
   bool  seek_done = FALSE;       /* was there a seek_label before exit? */
+  int	error_sync;		 /* error resync state */
 
   /* note the drill type to see if we need to make the user repeat */
   drill_type = SCR_COMMAND( line );
@@ -1005,7 +894,9 @@ do_speedtest( FILE *script, char *line ) {
       move( linenum, 0 );
       for ( p = data; *p == ASCII_SPACE && *p != ASCII_NULL; p++ )
 	ADDCH( *p );
-      for ( chars_typed = 0, errors = 0; *p != ASCII_NULL; p++ ) 
+
+      for ( chars_typed = 0, errors = 0, error_sync = 0;
+	    *p != ASCII_NULL; p++ ) 
 	{
 	  rc = getch_fl( (*p != ASCII_NL) ? *p : ASCII_SPACE );
 	  c = (char)rc;
@@ -1019,6 +910,7 @@ do_speedtest( FILE *script, char *line ) {
 	  if ( chars_typed == 0 )
 	    start_time = (long)time( NULL );
 	  chars_typed++;
+	  error_sync--;
       
 	  /* check for delete keys if not at line start or
 	     speed test start */
@@ -1046,12 +938,27 @@ do_speedtest( FILE *script, char *line ) {
 	    ADDCH( c );
 	  else 
 	    {
+	      /* try to sync with typist behind */
+	      if ( error_sync >= 0 && p > data && c == *(p-1) )
+		{
+		  p--;
+		  continue;
+		}
+
 	      ADDCH_REV( *p == ASCII_NL ?
 			 DRILL_NL_ERR : (unsigned char)*p );
 	      if ( ! cl_silent ) {
 		putchar( ASCII_BELL ); fflush( stdout );
 	      }
 	      errors++;
+	      error_sync = 1;
+
+	      /* try to sync with typist ahead */
+	      if ( c == *(p+1) )
+		{
+		  ungetch( c );
+		  error_sync++;
+		}
 	    }
       
 	  /* move screen location if newline */
@@ -1115,7 +1022,7 @@ do_speedtest( FILE *script, char *line ) {
 	      (float)errors/(float)chars_typed > global_error_max/100.0) 
 	    {
 	      sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
-	      wait_user( message, MODE_SPEEDTEST );
+	      wait_user (script, message, MODE_SPEEDTEST);
 
 	      /* check for F-command */
 	      if (global_on_failure_label != NULL)
@@ -1130,7 +1037,7 @@ do_speedtest( FILE *script, char *line ) {
 		  /* reset value unless persistent */
 		  if (!global_on_failure_label_persistent)
 		    global_on_failure_label = NULL;
-		  wait_user( message, MODE_SPEEDTEST );
+		  wait_user (script, message, MODE_SPEEDTEST);
 		  seek_done = TRUE;
 		  break;
 		}
@@ -1173,26 +1080,12 @@ do_speedtest( FILE *script, char *line ) {
 /*
  * clear the complete screen, maybe leaving a header behind
  */
-static void 
-do_clear( FILE *script, char *line ) {
-  int	colnum = 0;			/* loop counter */
-
+static void do_clear (FILE *script, char *line)
+{
   /* clear the complete screen */
   move( B_TOP_LINE , 0 ); clrtobot();
 
-  /* create an inverse video banner at the top of the screen */
-  for ( colnum = 0; colnum < COLS; colnum++ )
-    ADDCH_REV( ASCII_SPACE );
-
-  /* add any banner text */
-  move( B_TOP_LINE , 0 );
-  ADDSTR_REV( SCR_DATA( line ));
-
-  /* add the version information at top right */
-  move( B_TOP_LINE, COLS - (strlen(PACKAGE)+strlen( VERSION )+2) );
-  ADDSTR_REV( PACKAGE );
-  ADDSTR_REV( " " );
-  ADDSTR_REV( VERSION );
+  banner (SCR_DATA (line));
 
   /* finally, get the next script command */
   get_script_line( script, line );
@@ -1225,23 +1118,6 @@ do_goto( FILE *script, char *line, bool flag )
 }
 
 
-/*
-  exit from the program (implied on eof)
-*/
-static void 
-do_exit( FILE *script ) 
-{
-  
-  /* close up all files, reset the screen stuff, and exit */
-  fclose( script );
-  if ( cl_colour && has_colors() )
-    wbkgdset( stdscr, 0 );
-  clear(); refresh(); endwin();
-  printf( "\n" );
-  exit( 0 );
-}
-
-
 /*
   Ask the user whether he/she wants to repeat, continue or exit
   (this is used at the end of an exercise (drill/speedtest))
@@ -1278,8 +1154,7 @@ do_query_repeat ( FILE *script, bool allow_next )
 	break;
       }
       if (toupper ((char)resp) == 'E' || toupper ((char)resp) == RNE [4]) {
-	if (fkey_bindings [11] != NULL &&
-	    do_query_simple (CONFIRM_EXIT_LESSON_MSG))
+	if (do_query_simple (CONFIRM_EXIT_LESSON_MSG))
 	  {
 	    seek_label (script, fkey_bindings [11], NULL);
 	    resp = 'E';
@@ -1314,6 +1189,9 @@ do_query_simple ( char *text )
 {
   char resp;
 
+  if (user_is_always_sure)
+     return TRUE;
+  
   /* display the prompt */
   move( MESSAGE_LINE, 0 ); clrtoeol();
   move( MESSAGE_LINE, COLS - strlen( MODE_QUERY ) - 2 );
@@ -1620,9 +1498,12 @@ parse_file( FILE *script, char *label ) {
 	  do_speedtest( script, line ); break;
 	case C_KEYBIND:	do_keybind( script, line ); break;
 	  
-	case C_LABEL:	get_script_line( script, line ); break;
+	case C_LABEL:
+	   get_script_line (script, line);
+	   break;
 	case C_ERROR_MAX_SET: do_error_max_set( script, line ); break;
 	case C_ON_FAILURE_SET: do_on_failure_label_set( script, line ); break;
+	case C_MENU: do_menu (script, line); break;
 	default:
 	  fatal_error( _("unknown command"), line );
 	  break;
@@ -1715,8 +1596,11 @@ print_help()
       "-l L",
       "-w",
       "-k",
+      "-i",
       "-h",
-      "-v" };
+      "-v",
+      "-S",
+      "--banner-colors=F,B,P,V"};
   char *lop[]= 
     { "--max-error=%",
       "--notimer",
@@ -1728,8 +1612,11 @@ print_help()
       "--start-label=L",
       "--word-processor",
       "--no-skip",
+      "--show-errors",
       "--help",
-      "--version" };
+      "--version",
+      "--always-sure",
+      ""};
   char *help[] = 
     { 
       _("default maximum error percentage (default 3.0); valid values are between 0.0 and 100.0"),
@@ -1742,8 +1629,12 @@ print_help()
       _("start the lesson at label 'L'"),
       _("try to mimic word processors"),
       _("forbid the user to skip exercises"),
+      _("highlight errors with reverse video"),
       _("print this message"),
-      _("output version information and exit") };
+      _("output version information and exit"),
+      _("do not ask confirmation questions"),
+      _("set top banner colours (background, foreground, package and version "
+		      "respectively)")};
   
   int loop;
   
@@ -1788,17 +1679,21 @@ parse_cmdline( int argc, char **argv ) {
     { "curs-flash",	required_argument, 0, 'f' },
     { "colours",	required_argument, 0, 'c' },
     { "colors",		required_argument, 0, 'c' },
+    { "banner-colours",	required_argument, 0, '\x01' },
+    { "banner-colour",	required_argument, 0, '\x01' },
     { "silent",		no_argument, 0, 's' },
     { "quiet",		no_argument, 0, 'q' },
     { "start-label",	required_argument, 0, 'l' },
     { "word-processor",	no_argument, 0, 'w' },
     { "no-skip",        no_argument, 0, 'k' },
+    { "show-errors",    no_argument, 0, 'i' },
     { "help",		no_argument, 0, 'h' },
     { "version",	no_argument, 0, 'v' },
+    { "always-sure",	no_argument, 0, 'S' },
     { 0, 0, 0, 0 }};
 
   /* process every option */
-  while ( (c=getopt_long( argc, argv, "e:ntf:c:sql:wkhv",
+  while ( (c=getopt_long( argc, argv, "e:ntf:c:sql:wkihvS",
 			  long_options, &option_index )) != -1 ) 
     {
       switch (c)
@@ -1841,6 +1736,27 @@ parse_cmdline( int argc, char **argv ) {
 	    }
 	  cl_colour = TRUE;
 	  break;
+	case '\x01':
+	  if (sscanf (optarg, "%d,%d,%d,%d",
+			&cl_banner_bg_colour, &cl_banner_fg_colour,
+			&cl_prog_name_colour, &cl_prog_version_colour) != 4)
+	  {
+	     fprintf (stderr, _("%s:  argument format is incorrect\n"), optarg);
+	     exit (1);
+	  }
+
+	  if (cl_banner_bg_colour < 0 || cl_banner_bg_colour >= NUM_COLOURS ||
+		cl_banner_fg_colour < 0 || cl_banner_bg_colour >= NUM_COLOURS ||
+		cl_prog_version_colour < 0 ||
+		cl_prog_version_colour >= NUM_COLOURS ||
+		cl_prog_name_colour < 0 || cl_prog_name_colour >= NUM_COLOURS)
+	  {
+	     fprintf (stderr, _("%s:  incorrect colour values\n"), optarg);
+	     exit (1);
+	  }
+
+	  break;
+		
 	case 's':
 	case 'q':
 	  cl_silent = TRUE;
@@ -1860,6 +1776,9 @@ parse_cmdline( int argc, char **argv ) {
 	case 'k':
 	  cl_no_skip = TRUE;
 	  break;
+	case 'i':
+	  cl_rev_video_errors = TRUE;
+	  break;
 	case 'h':
 	  print_help();
 	  exit( 0 );
@@ -1870,6 +1789,9 @@ parse_cmdline( int argc, char **argv ) {
 	  printf( "%s\n", _("Written by Simon Baldwin"));
 	  exit( 0 );
 	  break;
+	case 'S':
+	  user_is_always_sure = TRUE;
+	  break;
 	case '?':
 	default:
 	  fprintf( stderr,
@@ -1907,7 +1829,6 @@ int main( int argc, char **argv )
 {
   WINDOW	*scr;				/* curses window */
   FILE	*script;			/* script file handle */
-  int	colnum = 0;			/* column counter */
   char	*p, filepath[FILENAME_MAX];	/* file paths */
   char	script_file[FILENAME_MAX];	/* more file paths */
   
@@ -1919,7 +1840,7 @@ int main( int argc, char **argv )
   textdomain (PACKAGE);
 #endif
 
-  COPYRIGHT=_("Copyright (C) 1998, 1999, 2000, 2001, 2002 Simon Baldwin.\n\
+  COPYRIGHT=_("Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Simon Baldwin.\n\
 This program comes with ABSOLUTELY NO WARRANTY; for details\n\
 please see the file 'COPYING' supplied with the source code.\n\
 This is free software, and you are welcome to redistribute it\n\
@@ -1933,29 +1854,30 @@ This program is released under the GNU General Public License.");
   MODE_DRILL=_("  Drill   ");
   /* this string is displayed in the mode-line when running a speedtest */
   MODE_SPEEDTEST=_("Speed test");
-  WAIT_MESSAGE=_(" Press Return to continue... ");
+  WAIT_MESSAGE=
+	  _(" Press RETURN or SPACE to continue, ESC to return to the menu ");
   /* this message is displayed when the user has failed in a [DS]: drill */
   ERROR_TOO_HIGH_MSG=
-    _("Your error-rate is too high. You have to achieve %.1f%%.");
+    _(" Your error-rate is too high. You have to achieve %.1f%%. ");
   /* this message is displayed when the user has failed in a [DS]: drill,
      and an F:<LABEL> ("on failure label") is in effect */
   SKIPBACK_VIA_F_MSG=
-    _("You failed this test, so you need to go back to %s.");
+    _(" You failed this test, so you need to go back to %s. ");
   /* this is used for repeat-queries. you can translate the keys as well
      (if you translate msgid "R/N/E" accordingly) */
   REPEAT_NEXT_EXIT_MSG= 
-	_("Press R to repeat, N for next exercise or E to exit");
+	_(" Press R to repeat, N for next exercise or E to exit ");
   /* this is used for repeat-queries with --no-skip. you can translate
      the keys as well (if you translate msgid "R/N/E" accordingly) */
   REPEAT_EXIT_MSG= 
-	_("Press R to repeat or E to exit");
+	_(" Press R to repeat or E to exit ");
   /* This is used make the user confirm (E)xit in REPEAT_NEXT_EXIT_MSG */
   CONFIRM_EXIT_LESSON_MSG=
-    _("Are you sure you want to exit this lesson? [Y/N]");
+    _(" Are you sure you want to exit this lesson? [Y/N] ");
   /* This message is displayed if the user tries to skip a lesson
      (ESC ESC) and --no-skip is specified */
   NO_SKIP_MSG= 
-    _("Sorry, gtypist is configured to forbid skipping exercises.");
+    _(" Sorry, gtypist is configured to forbid skipping exercises. ");
   /* this must be adjusted to the right with one space at the end.
      Leading whitespace is important because it is displayed in reverse
      video because it must be aligned with the next two messages 
@@ -2051,27 +1973,40 @@ This program is released under the GNU General Public License.");
   signal( SIGPIPE, catcher ); signal( SIGTERM, catcher );
   clear(); refresh(); typeahead( -1 );
   keypad( scr, TRUE ); noecho(); curs_set( 0 ); raw();
-  
+
+  // Quick hack to get rid of the escape delays
+#ifdef __NCURSES_H 
+  ESCDELAY = 1;
+#endif
+
   /* set up colour pairs if possible */
-  if ( cl_colour && has_colors() ) 
-    {
-      start_color();
-      init_pair( C_NORMAL,
-		 colour_array[ cl_fgcolour ],
-		 colour_array[ cl_bgcolour ] );
-      wbkgdset( stdscr, COLOR_PAIR( C_NORMAL ) );
-    }
+  if (has_colors ()) 
+  {
+     start_color ();
+
+     init_pair (C_NORMAL,
+		 colour_array [cl_fgcolour],
+		 colour_array [cl_bgcolour]);
+     wbkgdset (stdscr, COLOR_PAIR (C_NORMAL));
+
+     init_pair (C_BANNER,
+		     colour_array [cl_banner_fg_colour],
+		     colour_array [cl_banner_bg_colour]);
+     init_pair (C_PROG_NAME,
+		     colour_array [cl_banner_fg_colour],
+		     colour_array [cl_prog_name_colour]);
+     init_pair (C_PROG_VERSION,
+		     colour_array [cl_banner_fg_colour],
+		     colour_array [cl_prog_version_colour]);
+     init_pair (C_MENU_TITLE,
+		     colour_array [cl_menu_title_colour],
+		     colour_array [cl_bgcolour]);
+  }
   
   /* put up the top line banner */
   clear();
-  move( B_TOP_LINE , 0 );;
-  for ( colnum = 0; colnum < COLS; colnum++ )
-    ADDCH_REV( ASCII_SPACE );
-  move( B_TOP_LINE, COLS - strlen(PACKAGE) - strlen(VERSION)+2 );
-  ADDSTR_REV( PACKAGE );
-  ADDSTR_REV( " " );
-  ADDSTR_REV( VERSION );
-  
+  banner (_("Loading the script..."));
+ 
   /* index all the labels in the file */
   index_labels( script );
   
@@ -2082,3 +2017,4 @@ This program is released under the GNU General Public License.");
   /* for lint... */
   return( 0 );
 }
+
